from functools import cached_property, lru_cache
import glob
import json
import os.path
from pathlib import Path

import numpy as np
import pandas as pd
from scipy import signal

from abr.datatype import ABRWaveform, ABRSeries

from .dataset import DataCollection, Dataset


def get_filename(pathname, suffix='ABR average waveforms.csv'):
    if pathname.is_file():
        if pathname.name.endswith(suffix):
            return pathname
        else:
            raise IOError('Invalid ABR file')
    filename = pathname / suffix
    if filename.exists():
        return filename
    filename = pathname / f'{pathname.stem} {suffix}'
    if filename.exists():
        return filename
    raise IOError(f'Could not find average waveforms file for {pathname}')


@lru_cache(maxsize=64)
def read_file(filename):
    with filename.open() as fh:
        # This supports a variable-length header where we may not have included
        # some levels (e.g., epoch_n and epoch_reject_ratio).
        header = {}
        while True:
            line = fh.readline()
            if line.startswith('time'):
                break
            name, *keys = line.strip().split(',')
            keys = [-1 if k == 'click' else float(k) for k in keys]
            header[name] = np.array(keys).astype('f')
        data = pd.read_csv(fh, index_col=0, header=None)

    header = pd.MultiIndex.from_arrays(list(header.values()),
                                       names=list(header.keys()))
    data.index.name = 'time'
    data.index *= 1e3
    data.columns = header
    return data.T


class PSIDataCollection(DataCollection):

    def __init__(self, filename):
        filename = Path(filename)
        self.filename = get_filename(filename)

    @cached_property
    def fs(self):
        settings_file = get_filename(self.filename.parent, 'ABR processing settings.json')
        if settings_file.exists():
            fs = json.loads(settings_file.read_text())['actual_fs']
        else:
            fs = np.mean(np.diff(data.columns.values)**-1)
        return fs

    @cached_property
    def data(self):
        data = read_file(self.filename)
        keep = ['frequency', 'level']
        drop = [c for c in data.index.names if c not in keep]
        return data.reset_index(drop, drop=True)

    @cached_property
    def frequencies(self):
        return self.data.index.unique('frequency').values

    @property
    def name(self):
        return self.filename.parent.stem

    def iter_frequencies(self):
        for frequency in self.frequencies:
            yield PSIDataset(self, frequency)


class PSIDataset(Dataset):

    def __init__(self, parent, frequency):
        self.parent = parent
        self.frequency = frequency

    @property
    def filename(self):
        return self.parent.filename

    @property
    def fs(self):
        return self.parent.fs

    def get_series(self, filter_settings=None):
        data = self.parent.data.loc[self.frequency]
        if filter_settings is not None:
            Wn = filter_settings['highpass'], filter_settings['lowpass']
            N = filter_settings['order']
            b, a = signal.iirfilter(N, Wn, fs=self.fs)
            data_filt = signal.filtfilt(b, a, data.values, axis=-1)
            data = pd.DataFrame(data_filt, columns=data.columns, index=data.index)

        waveforms = []
        for level, w in data.iterrows():
            level = float(level)
            waveforms.append(ABRWaveform(self.fs, w, level))

        series = ABRSeries(waveforms, self.frequency)
        series.filename = self.parent.filename
        series.id = self.parent.filename.parent.name
        series.dataset = self
        return series


def iter_all(path):
    results = []
    path = Path(path)
    if 'abr_io' in path.stem:
        yield from PSIDataCollection(path).iter_frequencies()
    else:
        for subpath in path.glob('**/*abr_io'):
            yield from PSIDataCollection(subpath).iter_frequencies()
        for subpath in path.glob('**/*abr_io_click'):
            yield from PSIDataCollection(subpath).iter_frequencies()



import functools
from pathlib import Path
import re


P_RATER = re.compile('.*(?:kHz|click)(?:-(\w+))?-analyzed.txt')


def get_rater(filename):
    try:
        return P_RATER.match(filename.name).group(1)
    except:
        return 'Unknown'


class DataCollection:

    def __init__(self, filename):
        self.filename = Path(filename)

    @property
    def fs(self):
        raise NotImplementedError

    @property
    def data(self):
        raise NotImplementedError

    @property
    def frequencies(self):
        raise NotImplementedError

    @property
    def name(self):
        raise NotImplementedError

    def iter_frequencies(self):
        raise NotImplementedError


@functools.total_ordering
class Dataset:

    filename_template = '{filename}-{frequency}-{rater}analyzed.txt'

    def __init__(self, parent, frequency):
        self.parent = parent
        self.frequency = frequency

    @property
    def filename(self):
        return self.parent.filename

    @property
    def fs(self):
        return self.parent.fs

    def get_series(self, filter_settings=None):
        raise NotImplementedError

    def get_analyzed_filename(self, rater):
        if self.frequency == -1:
            frequency = 'click'
        else:
            frequency = round(self.frequency * 1e-3, 8)
            frequency = f'{frequency}kHz'

        filename = self.filename.with_suffix('')
        if rater != '*':
            rater = rater + '-'
        return self.filename_template \
            .format(filename=filename, frequency=frequency, rater=rater)

    def find_analyzed_files(self):
        glob_pattern = self.get_analyzed_filename('*')
        path = Path(glob_pattern)
        return list(path.parent.glob(path.name))

    def list_raters(self):
        return [get_rater(filename) for filename in self.find_analyzed_files()]

    def load_analysis(self, rater):
        from . import load_analysis
        return load_analysis(self.get_analyzed_filename(rater))

    def __lt__(self, other):
        if not isinstance(other, Dataset):
            raise NotImplemented
        return (self.filename, self.frequency) < \
            (other.filename, other.frequency)

    def __eq__(self, other):
        if not isinstance(other, Dataset):
            raise NotImplemented
        return (self.filename, self.frequency) == \
            (other.filename, other.frequency)

    def __hash__(self):
        return hash((self.filename, self.frequency))



import argparse

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from atom.api import Bool, Dict, Event, List, observe, Tuple, Typed, Str
import enaml
from enaml.application import deferred_call
from enaml.core.api import d_, Declarative
from enaml.qt.qt_application import QtApplication

with enaml.imports():
    from abr.compare_window import CompareWindow

from abr.presenter import WaveformPresenter


class Compare(Declarative):

    th = Typed(pd.Series)
    waves = Typed(pd.DataFrame)
    data = Typed(pd.DataFrame)

    as_difference = d_(Bool(False))
    jitter = d_(Bool(False))
    axes = Typed(plt.Axes)
    figure = Typed(plt.Figure)
    plot_map = Dict()

    rater_x = d_(Str())
    rater_y = d_(Str())
    selected_feature = d_(Str())
    selected_point = d_(Tuple(), writable=False)

    available_raters = List()
    available_features = List()
    available_subjects = List()

    def __init__(self, parser, directory):
        th, waves = parser.load_analyses(directory)
        ix = [c for c in waves.columns if not (c.endswith('Latency') or c.endswith('Amplitude') or 'msec' in c)]
        waves = waves.set_index(ix)
        ix.remove('Level')
        th = th.set_index(ix)['thresholds'].clip(lower=-20, upper=100)

        features = [c for c in waves.columns if 'msec' not in c.lower()]
        features.sort(key=lambda x: (int(x[1]), x[0] != 'P', x.split(' ')[1]))
        self.available_features = ['Threshold'] + features
        self.available_raters = th.index.unique('analyzer').tolist()

        super().__init__(th=th, waves=waves)
        self.request_update()

    def _observe_selected_feature(self, event):
        if self.selected_feature == 'Threshold':
            self.data = self.th.unstack('analyzer')
        else:
            self.data = self.waves[self.selected_feature].unstack('analyzer')

    def _default_figure(self):
        context = {
            'axes.spines.left': True,
            'ytick.left': True,
            'figure.subplot.left': 0.15,
            'figure.subplot.bottom': 0.1,
            'figure.subplot.top': 0.95,
            'figure.subplot.right': 0.95,
        }
        with plt.rc_context(context):
            figure, self.axes = plt.subplots(1, 1)
        return figure

    def _default_rater_x(self):
        return self.available_raters[0]

    def _default_rater_y(self):
        i = 1 if (len(self.available_raters) > 1) else 0
        return self.available_raters[i]

    def _default_selected_feature(self):
        return self.available_features[0]

    @observe('rater_x', 'rater_y', 'as_difference', 'jitter', 'selected_feature')
    def request_update(self, event=None):
        if event is not None and event['type'] == 'create':
            return
        deferred_call(self._update_plot)

    def _update_plot(self):
        self.axes.clear()
        self.plot_map = {}

        x = self.data.loc[:, self.rater_x].copy()
        y = self.data.loc[:, self.rater_y].copy()

        self.axes.set_xlabel(f'Rater {self.rater_x}')
        if self.as_difference:
            y -= x
            self.axes.set_ylabel(f'Difference between {self.rater_y} and {self.rater_x}')
        else:
            self.axes.set_ylabel(f'Rater {self.rater_y}')

        if self.jitter:
            bound = (x.max() - x.min()) * 0.025
            x += np.random.uniform(-bound, bound, len(x))
            y += np.random.uniform(-bound, bound, len(x))

        for dataset, _ in x.groupby('dataset'):
            xd = x.loc[dataset]
            yd = y.loc[dataset]
            l, = self.axes.plot(xd, yd, 'o', mec='w', mew=1)
            self.plot_map[l] = dataset

        if self.figure.canvas is not None:
            self.figure.canvas.draw()

    def button_press_event(self, event):
        if not event.inaxes:
            return
        for plot, dataset in self.plot_map.items():
            contained, info = plot.contains(event)
            if contained:
                if self.selected_feature == 'Threshold':
                    self.selected_point = dataset, None
                else:
                    level = self.data.loc[dataset].iloc[info['ind']].iloc[0].name
                    self.selected_point = dataset, level



import json
from pathlib import Path
from enaml.qt.QtCore import QStandardPaths


def config_path():
    config_path = Path(QStandardPaths.standardLocations(QStandardPaths.GenericConfigLocation)[0])
    return config_path / 'OHSU' / 'abr'


def config_file():
    config_file =  config_path() / 'config.json'
    config_file.parent.mkdir(exist_ok=True, parents=True)
    return config_file


def read_config():
    filename = config_file()
    if not filename.exists():
        return {}
    return json.loads(filename.read_text())


def write_config(config):
    filename = config_file()
    filename.write_text(json.dumps(config, indent=2))



import argparse

import enaml
from enaml.application import deferred_call
from enaml.qt.qt_application import QtApplication
from scipy import stats

with enaml.imports():
    from abr.launch_window import LaunchWindow, STORE
    from abr.main_window import (DNDWindow, load_files, SerialWindow)
    from abr.compare_window import CompareWindow
    from abr.presenter import SerialWaveformPresenter, WaveformPresenter


from abr.compare import Compare
from abr.parsers import Parser


P_LATENCIES = {
    1: stats.norm(1.5, 0.5),
    2: stats.norm(2.5, 1),
    3: stats.norm(3.0, 1),
    4: stats.norm(4.0, 1),
    5: stats.norm(5.0, 2),
}


def add_default_arguments(parser, waves=True):
    parser.add_argument('--nofilter', action='store_false', dest='filter',
                        default=True, help='Do not filter waveform')
    parser.add_argument('--lowpass',
                        help='Lowpass cutoff (Hz), default 3000 Hz',
                        default=3000, type=float)
    parser.add_argument('--highpass',
                        help='Highpass cutoff (Hz), default 300 Hz',
                        default=300, type=float)
    parser.add_argument('--order',
                        help='Filter order, default 1st order', default=1,
                        type=int)
    parser.add_argument('--parser', default='EPL', help='Parser to use')
    parser.add_argument('--user', help='Name of person analyzing data')
    if waves:
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument('--threshold-only', action='store_true')
        group.add_argument('--all-waves', action='store_true')
        group.add_argument('--waves', type=int, nargs='+')


def parse_args(parser, waves=True):
    options = parser.parse_args()
    exclude = ('filter', 'lowpass', 'highpass', 'order', 'parser', 'user',
               'waves', 'all_waves', 'threshold_only')
    new_options = {k: v for k, v in vars(options).items() if k not in exclude}
    filter_settings = None
    if options.filter:
        filter_settings = {
            'lowpass': options.lowpass,
            'highpass': options.highpass,
            'order': options.order,
        }
    new_options['parser'] = Parser(options.parser, filter_settings,
                                   options.user)

    if not waves:
        return new_options

    if options.all_waves:
        waves = [1, 2, 3, 4, 5]
    elif options.threshold_only:
        waves = []
    else:
        waves = options.waves[:]
    new_options['latencies'] = {w: P_LATENCIES[w] for w in waves}
    return new_options


def main():
    parser = argparse.ArgumentParser('abr')
    parser.add_argument('--clear-settings', action='store_true',
                        help='Clear persisted settings')
    args = parser.parse_args()
    if args.clear_settings:
        STORE.clear()

    app = QtApplication()
    window = LaunchWindow()
    window.show()
    app.start()
    app.stop()


def main_gui():
    parser = argparse.ArgumentParser('abr-gui')
    add_default_arguments(parser)
    parser.add_argument('--demo', action='store_true', dest='demo',
                        default=False, help='Load demo data')
    parser.add_argument('filenames', nargs='*')
    options = parse_args(parser)

    app = QtApplication()
    view = DNDWindow(parser=options['parser'], latencies=options['latencies'])
    deferred_call(load_files, options['parser'], options['latencies'],
                  options['filenames'], view.find('dock_area'))

    view.show()
    app.start()
    app.stop()


def main_batch():
    parser = argparse.ArgumentParser('abr-batch')
    add_default_arguments(parser)
    parser.add_argument('dirnames', nargs='*')
    parser.add_argument('--skip-errors', action='store_true')
    options = parse_args(parser)
    parser = options['parser']

    app = QtApplication()
    presenter = SerialWaveformPresenter(parser=parser,
                                        latencies=options['latencies'],
                                        paths=options['dirnames'])
    view = SerialWindow(presenter=presenter)
    view.show()
    app.start()
    app.stop()


def main_compare():
    parser = argparse.ArgumentParser("abr-compare")
    add_default_arguments(parser)
    parser.add_argument('directory')
    options = parse_args(parser)

    presenter_a = WaveformPresenter(latencies=options['latencies'], parser=options['parser'], interactive=False)
    presenter_b = WaveformPresenter(latencies=options['latencies'], parser=options['parser'], interactive=False)
    presenter_c = WaveformPresenter(latencies=options['latencies'], parser=options['parser'])

    app = QtApplication()
    compare = Compare(options['parser'], options['directory'])
    view = CompareWindow(compare=compare,
                         rater=options['parser']._user,
                         presenter_a=presenter_a,
                         presenter_b=presenter_b,
                         presenter_c=presenter_c,
                         )
    view.show()
    app.start()
    app.stop()



import threading
import queue

import numpy as np
import pandas as pd

from atom.api import (Atom, Typed, Dict, List, Bool, Int, Float, Tuple,
                      Property, Value, set_default)
from enaml.application import timed_call

from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib import transforms as T

from abr.abrpanel import WaveformPlot
from abr.datatype import ABRSeries, WaveformPoint, Point
from abr.parsers.dataset import Dataset


def plot_model(axes, model):
    n = len(model.waveforms)
    offset_step = 1/(n+1)
    plots = []

    text_trans = T.blended_transform_factory(axes.figure.transFigure,
                                             axes.transAxes)


    limits = [(w.y.min(), w.y.max()) for w in model.waveforms]
    base_scale = np.mean(np.abs(np.array(limits)))

    bscale_in_box = T.Bbox([[0, -base_scale], [1, base_scale]])
    bscale_out_box = T.Bbox([[0, -1], [1, 1]])
    bscale_in = T.BboxTransformFrom(bscale_in_box)
    bscale_out = T.BboxTransformTo(bscale_out_box)

    tscale_in_box = T.Bbox([[0, -1], [1, 1]])
    tscale_out_box = T.Bbox([[0, 0], [1, offset_step]])
    tscale_in = T.BboxTransformFrom(tscale_in_box)
    tscale_out = T.BboxTransformTo(tscale_out_box)

    minmax_in_box = T.Bbox([[0, 0], [1, 1]])
    minmax_out_box = T.Bbox([[0, 0], [1, 1]])
    minmax_in = T.BboxTransformFrom(minmax_in_box)
    minmax_out = T.BboxTransformTo(minmax_out_box)

    boxes = {
        'tscale': tscale_in_box,
        'tnorm': [],
        'norm_limits': limits/base_scale,
        'minmax': minmax_out_box,
    }

    for i, waveform in enumerate(model.waveforms):
        y_min, y_max = waveform.y.min(), waveform.y.max()
        tnorm_in_box = T.Bbox([[0, -1], [1, 1]])
        tnorm_out_box = T.Bbox([[0, -1], [1, 1]])
        tnorm_in = T.BboxTransformFrom(tnorm_in_box)
        tnorm_out = T.BboxTransformTo(tnorm_out_box)
        boxes['tnorm'].append(tnorm_in_box)

        offset = offset_step * i + offset_step * 0.5
        translate = T.Affine2D().translate(0, offset)

        y_trans = bscale_in + bscale_out + \
            tnorm_in + tnorm_out + \
            tscale_in + tscale_out + \
            translate + \
            minmax_in + minmax_out + \
            axes.transAxes
        trans = T.blended_transform_factory(axes.transData, y_trans)

        plot = WaveformPlot(waveform, axes, trans)
        plots.append(plot)

        text_trans = T.blended_transform_factory(axes.transAxes, y_trans)
        axes.text(-0.05, 0, f'{waveform.level}', transform=text_trans)

    axes.set_yticks([])
    axes.grid()
    for spine in ('top', 'left', 'right'):
        axes.spines[spine].set_visible(False)

    return plots, boxes


class WaveformPresenter(Atom):

    figure = Typed(Figure, {})
    raters = List()

    axes = Typed(Axes)
    dataset = Typed(Dataset)
    model = Typed(ABRSeries)

    current = Property()

    toggle = Property()
    scale = Property()
    normalized = Property()
    top = Property()
    bottom = Property()
    boxes = Dict()

    _current = Int()
    _toggle = Value()
    plots = List()

    threshold_marked = Bool(False)
    peaks_marked = Bool(False)
    valleys_marked = Bool(False)

    parser = Value()
    latencies = Dict()

    batch_mode = Bool(False)
    interactive = Bool(True)
    modified = Bool(False)

    def _default_axes(self):
        axes = self.figure.add_axes([0.1, 0.1, 0.8, 0.8])
        return axes

    def __init__(self, parser, latencies, interactive=True):
        self.parser = parser
        self.latencies = latencies
        self.interactive = interactive

    def load(self, dataset):
        self.dataset = dataset
        self.raters = dataset.list_raters()

        self._current = 0
        self.axes.clear()
        self.axes.set_xlabel('Time (msec)')
        self.model = self.parser.load(dataset)
        self.plots, self.boxes = plot_model(self.axes, self.model)

        self.normalized = False
        self.threshold_marked = False
        self.peaks_marked = False
        self.valleys_marked = False

        # Set current before toggle. Ordering is important.
        self.current = len(self.model.waveforms)-1
        self.toggle = None
        self.update()
        self.modified = False

    def save(self):
        if np.isnan(self.model.threshold):
            raise ValueError('Threshold not set')
        if self.latencies:
            if not self.peaks_marked or not self.valleys_marked:
                raise ValueError('Waves not identified')
        self.parser.save(self.model)
        self.raters = self.dataset.list_raters()
        self.modified = False

    def update(self):
        for p in self.plots:
            p.update()
        if self.axes.figure.canvas is not None:
            self.axes.figure.canvas.draw()

    def _get_current(self):
        return self._current

    def _set_current(self, value):
        if not (0 <= value < len(self.model.waveforms)):
            return
        if value == self.current:
            return
        self.plots[self.current].current = False
        self.plots[value].current = True
        self._current = value
        self.update()

    def _get_scale(self):
        return self.boxes['tscale'].ymax

    def _set_scale(self, value):
        if value < 0:
            return
        box = np.array([[0, -value], [1, value]])
        self.boxes['tscale'].set_points(box)
        self.update()

    def _get_normalized(self):
        box = self.boxes['tnorm'][0]
        return not ((box.ymin == -1) and (box.ymax == 1))

    def _set_normalized(self, value):
        if value:
            zipped = zip(self.boxes['tnorm'], self.boxes['norm_limits'])
            for box, (lb, ub) in zipped:
                points = np.array([[0, lb], [1, ub]])
                box.set_points(points)
        else:
            for box in self.boxes['tnorm']:
                points = np.array([[0, -1], [1, 1]])
                box.set_points(points)
        self.axes.set_title('normalized' if value else 'raw')
        self.update()

    def _get_top(self):
        return self.boxes['minmax'].ymax

    def _set_top(self, value):
        points = np.array([[0, self.bottom], [1, value]])
        self.boxes['minmax'].set_points(points)
        self.update()

    def _get_bottom(self):
        return self.boxes['minmax'].ymin

    def _set_bottom(self, value):
        points = np.array([[0, value], [1, self.top]])
        self.boxes['minmax'].set_points(points)
        self.update()

    def set_suprathreshold(self):
        self.set_threshold(-np.inf)

    def set_subthreshold(self):
        self.set_threshold(np.inf)
        if self.latencies:
            if not self.peaks_marked:
                self.guess()
            if not self.valleys_marked:
                self.guess()

    def set_threshold(self, threshold=None):
        if threshold is None:
            threshold = self.get_current_waveform().level
        self.model.threshold = threshold
        self.threshold_marked = True
        if self.latencies and not self.peaks_marked:
            self.guess()
        self.update()
        self.modified = True

    def _get_toggle(self):
        return self._toggle

    def _set_toggle(self, value):
        if value == self._toggle:
            return
        for plot in self.plots:
            point = plot.point_plots.get(self.toggle)
            if point is not None:
                point.current = False
        self._toggle = value
        for plot in self.plots:
            point = plot.point_plots.get(value)
            if point is not None:
                point.current = True
        self.update()

    def guess(self):
        if not self.latencies:
            return
        if not self.peaks_marked:
            self.model.guess_p(self.latencies)
            ptype = Point.PEAK
            self.peaks_marked = True
        elif not self.valleys_marked:
            self.model.guess_n()
            ptype = Point.VALLEY
            self.valleys_marked = True
        else:
            return
        self.update()
        self.current = len(self.model.waveforms)-1
        self.toggle = 1, ptype
        self.update()
        self.modified = True

    def update_point(self):
        level = self.model.waveforms[self.current].level
        self.model.update_guess(level, self.toggle)
        self.update()

    def move_selected_point(self, step):
        point = self.get_current_point()
        point.move(step)
        self.update()

    def set_selected_point(self, time):
        try:
            point = self.get_current_point()
            index = point.time_to_index(time)
            point.move(('set', index))
            self.update()
            self.modified = True
        except:
            pass

    def toggle_selected_point_unscorable(self):
        try:
            point = self.get_current_point()
            point.unscorable = not point.unscorable
            self.update()
            self.modified = True
        except:
            pass

    def mark_unscorable(self, mode):
        try:
            for waveform in self.model.waveforms:
                if mode == 'all':
                    waveform.points[self.toggle].unscorable = True
                elif mode == 'descending':
                    if waveform.level <= self.get_current_waveform().level:
                        waveform.points[self.toggle].unscorable = True
            self.update()
            self.modified = True
        except:
            pass

    def get_current_waveform(self):
        return self.model.waveforms[self.current]

    def get_current_point(self):
        return self.get_current_waveform().points[self.toggle]

    def clear_points(self):
        self.model.clear_points()
        self.peaks_marked = False
        self.valleys_marked = False
        self.update()
        self.modified = True

    def clear_peaks(self):
        self.model.clear_peaks()
        self.peaks_marked = False
        self.update()
        self.modified = True

    def clear_valleys(self):
        self.model.clear_valleys()
        self.valleys_marked = False
        self.update()
        self.modified = True

    def load_analysis(self, rater):
        self.clear_points()
        _, th, peaks = self.dataset.load_analysis(rater)
        self.model.load_analysis(th, peaks)
        self.peaks_marked = True
        self.valleys_marked = True
        self.update()
        self.modified = True

    def select_waveform(self, level):
        for i, waveform in enumerate(self.model.waveforms):
            if waveform.level == level:
                self.current = i

    def select_point(self, point):
        number = int(point[1])
        if point[0] == 'P':
            self.toggle = number, Point.PEAK
        else:
            self.toggle = number, Point.VALLEY


def scan_worker(parser, paths, queue, stop):
    for path in paths:
        for ds in parser.find_unprocessed(path):
            queue.put(('append', ds))
            if stop.is_set():
                break
        if stop.is_set():
            break
    queue.put(('complete',))


class SerialWaveformPresenter(WaveformPresenter):

    unprocessed = List()
    n_unprocessed = Int(0)

    current_model = Int(-1)
    batch_mode = set_default(True)

    scan_paths = Value()
    scan_queue = Value()
    scan_stop_event = Value()
    scan_complete = Bool(False)
    scan_thread = Value()

    def __init__(self, parser, latencies, paths):
        super().__init__(parser, latencies)
        self.scan_paths = paths
        self.scan_queue = queue.Queue()
        self.scan_stop_event = threading.Event()
        args = (self.parser, self.scan_paths, self.scan_queue, self.scan_stop_event)
        self.scan_thread = threading.Thread(target=scan_worker, args=args)
        self.scan_thread.start()
        self.unprocessed = []
        timed_call(100, self.scan_poll)

    def scan_poll(self):
        while True:
            try:
                mesg = self.scan_queue.get(block=False)
                if mesg[0] == 'append':
                    self.unprocessed.append(mesg[1])
                elif mesg[0] == 'complete':
                    self.scan_complete = True
            except queue.Empty:
                if not self.scan_complete:
                    timed_call(100, self.scan_poll)
                break
        self.n_unprocessed = len(self.unprocessed)
        if self.current_model < 0:
            self.load_next()

    def scan_stop(self):
        self.scan_stop_event.set()

    def load_model(self):
        fs = self.unprocessed[self.current_model]
        self.load(fs)

    def load_prior(self):
        if self.current_model < 1:
            return
        self.current_model -= 1
        self.load_model()

    def load_next(self):
        if self.current_model >= (len(self.unprocessed) - 1):
            return
        self.current_model += 1
        self.load_model()

    def save(self):
        super().save()
        self.load_next()



'''
This module defines the import/export routines for interacting with the data
store.  If you wish to customize this, simply define the load() function.

load(run_location, invert, filter) -- When the program needs a run loaded, it
will pass the run_location provided by list().  Invert is a boolean flag
indicating whether waveform polarity should be flipped.  Filter is a dictionary
containing the following keys:
    1. ftype: any of None, butterworth, bessel, etc.
    2. fh: highpass cutoff (integer in Hz)
    3. fl: lowpass cutoff (integer in Hz)
All objects of the epl.datatype.Waveform class will accept the filter
dictionary and perform the appropriate filtering.  It is recommended you use the
filtering provided by the Waveform class as the parameters of the filter will
also be recorded.  This function must return an object of the
epl.datatype.ABRSeries class.  See this class for appropriate documentation.

The save function must return a message.  If there is an error in saving, throw
the appropriate exception.
'''

import importlib
import re
from glob import glob
import os
from pathlib import Path
import time

import pandas as pd
import numpy as np

import abr
from ..datatype import Point


def waveform_string(waveform):
    data = [f'{waveform.level:.2f}']
    data.append(f'{waveform.mean(0, 1)}')
    data.append(f'{waveform.std(0, 1)}')
    for _, point in sorted(waveform.points.items()):
        data.append(f'{point.latency:.8f}')
        data.append(f'{point.amplitude:.8f}')
    return '\t'.join(data)


def filter_string(waveform):
    if getattr(waveform, '_zpk', None) is None:
        return 'No filtering'
    t = 'Pass %d -- z: %r, p: %r, k: %r'
    filt = [t % (i, z, p, k) for i, (z, p, k) in enumerate(waveform._zpk)]
    return '\n' + '\n'.join(filt)


def load_analysis(fname):
    th_match = re.compile('Threshold \(dB SPL\): ([\w.-]+)')
    freq_match = re.compile('Frequency \(kHz\): ([\d.]+)|Stimulus: ([.\w]+)?')
    with open(fname) as fh:
        text = fh.readline()
        th = th_match.search(text).group(1)
        th = None if th == 'None' else float(th)
        text = fh.readline()

        match = freq_match.search(text)
        if match.group(1) is not None:
            freq = float(match.group(1))
        else:
            freq = match.group(2)
            if freq == 'click':
                freq = -1
            else:
                freq = float(freq)

        for line in fh:
            if line.startswith('NOTE'):
                break
        data = pd.io.parsers.read_csv(fh, sep='\t', index_col='Level')
    return (freq, th, data)


def parse_peaks(peaks, threshold):
    # Convert the peaks dataframe to a format that can be used by _set_points.
    p_pattern = re.compile('P(\d) Latency')
    n_pattern = re.compile('N(\d) Latency')

    p_latencies = {}
    n_latencies = {}

    for c in peaks:
        match = p_pattern.match(c)
        if match:
            wave = int(match.group(1))
            p_latencies[wave] = pd.DataFrame({'x': peaks[c]})
        match = n_pattern.match(c)
        if match:
            wave = int(match.group(1))
            n_latencies[wave] = pd.DataFrame({'x': peaks[c]})

    p_latencies = pd.concat(p_latencies.values(), keys=p_latencies.keys(),
                            names=['wave'])
    p_latencies = {g: df.reset_index('Level', drop=True) \
                   for g, df in p_latencies.groupby('Level')}
    n_latencies = pd.concat(n_latencies.values(), keys=n_latencies.keys(),
                            names=['wave'])
    n_latencies = {g: df.reset_index('Level', drop=True) \
                   for g, df in n_latencies.groupby('Level')}

    for level, df in p_latencies.items():
        if level < threshold:
            df[:] = -df[:]

    for level, df in n_latencies.items():
        if level < threshold:
            df[:] = -df[:]

    return p_latencies, n_latencies


class Parser(object):

    def __init__(self, file_format, filter_settings, user=None):
        '''
        Parameters
        ----------
        file_format : string
            File format that will be loaded.
        filter_settings : {None, dict}
            If None, no filtering is applied. If dict, must contain ftype,
            lowpass, highpass and order as keys.
        user : {None, string}
            Person analyzing the data.
        '''
        self._file_format = file_format
        self._filter_settings = filter_settings
        self._rater = user
        self._module_name = f'abr.parsers.{file_format}'
        self._module = importlib.import_module(self._module_name)

    def load(self, fs):
        return fs.get_series(self._filter_settings)

    def save(self, model):
        # Assume that all waveforms were filtered identically
        filter_history = filter_string(model.waveforms[-1])

        # Generate list of columns
        columns = ['Level', '1msec Avg', '1msec StDev']
        point_keys = sorted(model.waveforms[0].points)
        for point_number, point_type in point_keys:
            point_type_code = 'P' if point_type == Point.PEAK else 'N'
            for measure in ('Latency', 'Amplitude'):
                columns.append(f'{point_type_code}{point_number} {measure}')

        columns = '\t'.join(columns)
        spreadsheet = '\n'.join(waveform_string(w) \
                                for w in reversed(model.waveforms))

        if model.freq == -1:
            stimulus = 'Stimulus: click'
        else:
            stimulus = f'Stimulus: {model.freq*1e-3:.2f} kHz'

        content = CONTENT.format(threshold=model.threshold,
                                 stimulus=stimulus,
                                 filter_history=filter_history,
                                 columns=columns,
                                 spreadsheet=spreadsheet,
                                 version=abr.__version__)


        filename = model.dataset.get_analyzed_filename(self._rater)
        with open(filename, 'w') as fh:
            fh.writelines(content)

    def iter_all(self, path):
        yield from self._module.iter_all(path)

    def find_processed(self, path):
        for ds in self.iter_all(path):
            if Path(ds.get_analyzed_filename(self._rater).exists()):
                yield ds

    def find_unprocessed(self, path):
        for ds in self.iter_all(path):
            if not Path(ds.get_analyzed_filename(self._rater)).exists():
                yield ds

    def find_analyses(self, study_directory):
        analyzed = {}
        for ds in self.iter_all(study_directory):
            analyzed[ds] = ds.find_analyzed_files()
        return analyzed

    def load_analyses(self, study_directory):
        keys = []
        thresholds = []
        waves = []
        analyses = self.find_analyses(study_directory)

        for ds, analyzed_filenames in analyses.items():
            for a in analyzed_filenames:
                _, th, w = load_analysis(a)
                parts = a.stem.split('-')
                if parts[-2].endswith('kHz'):
                    analyzer = 'Unknown'
                else:
                    analyzer = parts[-2]

                keys.append((ds, analyzer))
                thresholds.append(th)
                waves.append(w)

        names = ['dataset', 'analyzer']
        index = pd.MultiIndex.from_tuples(keys, names=names)
        thresholds = pd.Series(thresholds, index=index, name='thresholds').reset_index()
        waves = pd.concat(waves, keys=keys, names=names).reset_index()
        return thresholds, waves


CONTENT = '''
Threshold (dB SPL): {threshold:.2f}
{stimulus}
Filter history (zpk format): {filter_history}
file_format_version: 0.0.3
code_version: {version}
NOTE: Negative latencies indicate no peak. NaN for amplitudes indicate peak was unscorable.
{columns}
{spreadsheet}
'''.strip()


PARSER_MAP = {
    'PSI': 'psiexperiment',
    'NCRAR': 'IHS text export',
    'EPL': 'EPL CFTS',
}



# Use https://www.xiconeditor.com to convert to ico
import matplotlib.pyplot as plt
import matplotlib as mp
import numpy as np

fig = plt.figure(frameon=False)
fig.set_size_inches(2, 2)
ax = plt.Axes(fig, [0, 0, 1, 1])
ax.set_axis_off()
fig.add_axes(ax)

t = np.arange(-6, 6, 0.001)
i = 6
x = np.cos(t*1) * np.exp(-(t**2)/4)
o = np.exp(-i/10)
shift = 0
ax.fill_between(t+shift, x*o + 0.2, color='midnightblue')
ax.fill_between(t+shift, x*o + 0.2, 1, color='cornflowerblue')
ax.plot(t+shift, x*o + 0.2, color='white', lw=3)

ax.axis(xmin=-5, xmax=5, ymin=0, ymax=1)

patch = mp.patches.Rectangle([0, 0], width=1, height=1, facecolor='none',
                             edgecolor='white', linewidth=10,
                             transform=ax.transAxes)
ax.add_patch(patch)

fig.savefig('abr.png')
plt.show()
